from scrapemeagain.config import Config
from scrapemeagain.dockerized.utils import app_factory


app = app_factory(__name__)


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=Config.HEALTHCHECK_PORT)



from random import randint

from scrapemeagain.dockerized.ipchanger import DockerizedTorIpChanger


class FakeDockerizedTorIpChanger(DockerizedTorIpChanger):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._real_ip = "127.0.0.1"

    def get_current_ip(self):
        # Credits: https://stackoverflow.com/a/21014713/4183498.
        return ".".join(str(randint(0, 255)) for _ in range(4))

    def _obtain_new_ip(self):
        # We don't want to change IP address while testing.
        pass



import imp
import os

from tests import REPO_DIR


def import_docker_compose():
    module_name = "scrapemeagain-compose"
    module_path = os.path.join(REPO_DIR, "scripts")
    f, filename, description = imp.find_module(module_name, [module_path])

    return imp.load_module(module_name, f, filename, description)



"""Multiprocess I/O intensive tasks to maximize performance."""


import logging
from multiprocessing import Event, Pool, Process, Queue, Value
import time

from scrapemeagain.config import Config
from scrapemeagain.utils.alnum import get_current_datetime
from scrapemeagain.utils.http import get


EXIT = "__exit__"
DUMP_URLS_BUCKET = "__dump_urls_bucket__"


class Pipeline:
    def __init__(self, scraper, databaser, tor_ip_changer):
        """Webscraping pipeline.

        How it works:
        1. generate URLs for item list pages
        2. scrape item URLs from list pages
        3. store scraped item URLs in the DB
        4. load item URLs from the DB
        5. collect item properties
        6. store collected data in the DB

        :argument scraper: a scraper instance
        :type scraper: object
        :argument databaser: a databaser instance
        :type databaser: object
        :argument tor_ip_changer: a TorIpChanger instance
        :type tor_ip_changer: object
        """
        self.scraper = scraper
        self.databaser = databaser
        self.tor_ip_changer = tor_ip_changer

        self.scrape_processes = Config.SCRAPE_PROCESSES

        self.workers = []

    def prepare_multiprocessing(self):
        """Prepare all necessary multiprocessing objects."""
        self.url_queue = Queue()
        self.response_queue = Queue()
        self.data_queue = Queue()

        self.pool = Pool(self.scrape_processes)

        self.producing_urls_in_progress = Event()
        self.requesting_in_progress = Event()
        self.scraping_in_progress = Event()

        self.urls_to_process = Value("i", 0)
        self.urls_processed = Value("i", 0)
        self.urls_bucket_empty = Value("i", 1)

    def inform(self, message, log=True, end="\n"):
        """Print and if set log a message.

        :argument message:
        :type message: str
        :argument log: flag to also log the message
        :type log: bool
        :argument end: message line end
        :type end: str
        """
        if log:
            logging.info(message)

        print("{0} {1}".format(get_current_datetime(), message), end=end)

    def _inform_progress(self):
        """Print a message about how the scraping is progressing."""
        try:
            message = "Processed {0} ({1:.2f}%) URLs".format(
                self.urls_processed.value,
                self.urls_processed.value / self.urls_to_process.value * 100,
            )
            self.inform(message, log=False, end="\r")
        except ZeroDivisionError:
            pass

    def change_ip(self):
        """Change IP address.

        By default, IP is changed after each bunch of URLs is requested.
        """
        try:
            new_ip = self.tor_ip_changer.get_new_ip()
            logging.info("New IP: {new_ip}".format(new_ip=new_ip))
        except:
            logging.error("Failed setting new IP")
            return self.change_ip()

    def _produce_urls_wrapper(self, producer_function):
        """Use the provided 'producer_function' to populate 'url_queue'
        and get URLs count.

        :argument producer_function: how to produce URLs
        :type producer_function: function/method
        """
        self.producing_urls_in_progress.set()

        urls_count = producer_function()

        self.inform("URLs to process: {}".format(urls_count))
        time.sleep(1)
        self.urls_to_process.value = urls_count

        self.producing_urls_in_progress.clear()

    def _actually_produce_list_urls(self):
        return self.scraper.generate_list_urls()

    def produce_list_urls(self):
        """Populate 'url_queue' with generated list URLs."""

        def producer_function():
            urls_count = 0
            for list_url in self._actually_produce_list_urls():
                self.url_queue.put(list_url)
                urls_count += 1

            return urls_count

        self._produce_urls_wrapper(producer_function)

    def _actually_produce_item_urls(self):
        return self.databaser.get_item_urls()

    def produce_item_urls(self):
        """Populate 'url_queue' with loaded item URLs."""

        def producer_function():
            urls_count = 0
            for item_url in self._actually_produce_item_urls():
                # Pick the URL from SQLAlchemy ResultProxy.
                self.url_queue.put(item_url[0])
                urls_count += 1

            return urls_count

        self._produce_urls_wrapper(producer_function)

    def _classify_response(self, response):
        """Examine response and put it to 'response_queue' if it's OK or put
        it's URL  back to 'url_queue'.

        :argument response:
        :type response: request.response
        """
        if not response.ok and response.status_code >= 408:
            self.url_queue.put(response.url)
        else:
            self.response_queue.put(response)

    def _actually_get_html(self, urls):
        """Request provided URLs running multiple processes.

        :argument urls: URLs to get data from
        :type urls: list
        """
        try:
            self.requesting_in_progress.set()
            for response in self.pool.map(get, urls):
                self._classify_response(response)
        except Exception as exc:
            logging.error("Failed scraping URLs")
            logging.exception(exc)
        finally:
            self.requesting_in_progress.clear()

    def get_html(self):
        """Get HTML for URLs from 'url_queue'."""
        run = True

        while run:
            urls_bucket = []
            self.urls_bucket_empty.value = 1
            for _ in range(0, self.scrape_processes):
                url = self.url_queue.get()

                if url == EXIT:
                    run = False
                    break
                elif url == DUMP_URLS_BUCKET:
                    break

                urls_bucket.append(url)
                if self.urls_bucket_empty.value:
                    self.urls_bucket_empty.value = 0

            if urls_bucket:
                self._actually_get_html(urls_bucket)

            if run:
                self.change_ip()

    def _scrape_data(self, response):
        """Scrape HTML provided by the given response.

        :argument response:
        :type response: request.response

        :returns dict
        """
        if self.scraper.list_url_template in response.url:
            data = self.scraper.get_item_urls(response)
        else:
            data = self.scraper.get_item_properties(response)

        return data

    def _actually_collect_data(self, response):
        """Collect data from the given response.

        :argument response:
        :type response: request.response
        """
        try:
            self.scraping_in_progress.set()

            data = self._scrape_data(response)
            if data:
                self.data_queue.put(data)
        except Exception as exc:
            logging.error(
                'Failed processing response for "{}"'.format(response.url)
            )
            logging.exception(exc)
        finally:
            self.scraping_in_progress.clear()

    def collect_data(self):
        """Get data for responses from 'response_queue'."""
        while True:
            response = self.response_queue.get()

            if response == EXIT:
                break

            self._actually_collect_data(response)

    def _store_item_urls(self, data):
        """Handle storing item URLs.

        :argument data: item URLs
        :type data: list
        """
        if not data:
            return

        self.databaser.insert_multiple(data, self.databaser.item_urls_table)

    def _store_item_properties(self, data):
        """Handle storing item properties.

        :argument data: item properties
        :type data: dict
        """
        if len(data) > 1:
            # NOTE: if there is only a single item in the data dict (the URL),
            # there is no point in storing it.
            self.databaser.insert(data, self.databaser.item_data_table)

        # Remove processed item URL.
        self.databaser.delete_url(data["url"])

    def _actually_store_data(self, data):
        """Store provided data in the DB.

        :argument data: data to store in the DB
        :type data: str or list or dict
        """
        try:
            if isinstance(data, list):
                self._store_item_urls(data)
            else:
                self._store_item_properties(data)
        except Exception as exc:
            logging.error("Failed storing data")
            logging.exception(exc)
        finally:
            self.urls_processed.value += 1

    def store_data(self):
        """Consume 'data_queue' and store provided data in the DB."""
        self.urls_processed.value = 0

        while True:
            data = self.data_queue.get()

            if data == EXIT:
                break

            self._actually_store_data(data)

        self.databaser.commit()

    def exit_workers(self):
        """Exit workers started as separate processes by passing an EXIT
        message to all queues. This action leads to exiting `while` loops which
        workers processes run in.
        """
        self.inform("Exiting workers, please wait ...")

        self.url_queue.put(EXIT)
        self.response_queue.put(EXIT)
        self.data_queue.put(EXIT)

    def _queues_empty(self):
        """Check if queues are empty.

        :returns bool
        """
        return (
            self.url_queue.empty()
            and self.response_queue.empty()
            and self.data_queue.empty()
        )

    def _workers_idle(self):
        """Check if workers are idle.

        :returns bool
        """
        return (
            not self.producing_urls_in_progress.is_set()
            and not self.requesting_in_progress.is_set()
            and not self.scraping_in_progress.is_set()
        )

    def switch_power(self):
        """Check when to exit workers so the program won't run forever."""
        while True:
            # Check if workers can end.
            if (
                self._queues_empty()
                and self._workers_idle()
                and self.urls_bucket_empty.value
            ):
                self.exit_workers()
                break

            # Ensure all URLs are processed.
            if (
                self._queues_empty()
                and self._workers_idle()
                and not self.urls_bucket_empty.value
            ):
                logging.info("Dumping URLs bucket")
                self.url_queue.put(DUMP_URLS_BUCKET)

            # Inform about the progress.
            self._inform_progress()
            time.sleep(5)

    def employ_worker(self, target):
        """Create and register a daemon worker process.

        :argument target: worker's task
        :type target: function
        """
        worker = Process(target=target)
        worker.daemon = True
        worker.start()

        self.workers.append(worker)

    def release_workers(self):
        """Wait till all worker daemons are finished."""
        for worker in self.workers:
            worker.join()

    def get_item_urls(self):
        """Get item URLs from item list pages."""
        self.inform("Collecting item URLs")

        # scraper --> url_queue.
        self.employ_worker(self.produce_list_urls)

        # response_queue --> data_queue.
        self.employ_worker(self.collect_data)

        # data_queue --> DB.
        self.employ_worker(self.store_data)

        # Prevent running forever.
        self.employ_worker(self.switch_power)

        # NOTE Execution will block until 'get_html' is finished.
        # url_queue --> response_queue.
        self.get_html()

        self.release_workers()

    def get_item_properties(self):
        """Get item properties from item pages."""
        self.inform("Collecting item properties")

        # DB --> url_queue.
        self.employ_worker(self.produce_item_urls)

        # response_queue --> data_queue.
        self.employ_worker(self.collect_data)

        # data_queue --> DB.
        self.employ_worker(self.store_data)

        # Prevent running forever.
        self.employ_worker(self.switch_power)

        # NOTE Execution will block until 'get_html' is finished.
        # url_queue --> response_queue.
        self.get_html()

        self.release_workers()


class DockerizedPipeline(Pipeline):
    def inform(self, message, log=True, **kwargs):
        # Prevent using `end = '\r'` as that way docker-compose won't show all
        # messages.
        super().inform(message, log=log, end="\n")



"""Common HTTP functions."""


import logging
from random import sample

import requests

from scrapemeagain.config import Config


RESPONSE_LOG_MESSAGE = "{status} - {url}"


def get(url, **kwargs):
    """GET data from provided URL.

    :argument url:
    :type url: str

    :returns `requests.Response` instance
    """
    kwargs["proxies"] = {
        "http": Config.LOCAL_HTTP_PROXY,
        "https": Config.LOCAL_HTTP_PROXY,
    }

    kwargs["verify"] = False
    kwargs["timeout"] = Config.REQUEST_TIMEOUT

    user_agent = sample(Config.USER_AGENTS, 1)[0]
    kwargs["headers"] = {"User-Agent": user_agent}

    try:
        response = requests.get(url, **kwargs)

        logging.debug(
            RESPONSE_LOG_MESSAGE.format(status=response.status_code, url=url)
        )

        # NOTE: set the actually requested URL (as that one is stored in the
        # item_urls DB table).
        if url != response.url:
            logging.warning("Requested {0} got {1}".format(url, response.url))
            response.url = url
    except Exception as exc:
        # Don't fail on any exception and setup a fake response instead.
        response = requests.Response()
        response.url = url

        if isinstance(exc, requests.exceptions.Timeout):
            response.status_code = 408
        else:
            response.status_code = 503

        error_message = RESPONSE_LOG_MESSAGE.format(
            status=response.status_code, url=url
        )

        try:
            error_message += " - {}".format(str(exc))
        finally:
            logging.error(error_message)

    return response



from datetime import datetime
import json
import os

from requests import get
from bs4 import BeautifulSoup

from .alnum import DATE_FORMAT, get_current_date


USER_AGENTS_URL = (
    "https://techblog.willshouse.com/2012/01/03/most-common-user-agents/"
)
USER_AGENTS_FILE = os.path.join(
    os.path.dirname(os.path.dirname(os.path.dirname(__file__))),
    "useragents.json",
)


def _scrape_user_agents():
    response = get(USER_AGENTS_URL)
    soup = BeautifulSoup(response.content, "html.parser")

    user_agents = soup.findAll("td", {"class": "useragent"})
    user_agents = [user_agent.text for user_agent in user_agents]

    if not user_agents:
        raise ValueError("No user agents found!")

    return user_agents


def _save_user_agents(user_agents):
    data = {"date": get_current_date(), "useragents": user_agents}

    with open(USER_AGENTS_FILE, "w") as f:
        json.dump(data, f, indent=2)


def _scrape_and_save_user_agents():
    user_agents = _scrape_user_agents()
    _save_user_agents(user_agents)

    return user_agents


def _user_agents_are_old(user_agents):
    scrape_date = datetime.strptime(user_agents.get("date"), DATE_FORMAT)
    today = datetime.now()

    delta = today - scrape_date

    return delta.days > 30


def get_user_agents():
    if not os.path.exists(USER_AGENTS_FILE):
        return _scrape_and_save_user_agents()

    with open(USER_AGENTS_FILE, "r") as f:
        user_agents = json.load(f)

    if _user_agents_are_old(user_agents):
        return _scrape_and_save_user_agents()

    return set(user_agents["useragents"])



import unittest
from unittest.mock import patch

from tests.utils import import_docker_compose


docker_compose = import_docker_compose()


EXPECTED_EXAMPLESCRAPER_COMPOSE_DICT = {
    "version": "3",
    "services": {
        "examplescraper-scp1": {
            "environment": [
                "SERVICE_NAME_TEMPLATE=examplescraper-scp{}",
                "SERVICE_NAME_MASTER_SCRAPER=examplescraper-scp1",
                "DOCKER_HOST_IP=fake_docker_host_ip",
                "TOR_PORT=9051",
                "TOR_PASSWORD=I-solemnly-swear-I-am-up-to-no-good",
                "PRIVOXY_PORT=8118",
                "PRIVOXY_HOST=127.0.0.1",
                "IPSTORE_PORT=5000",
                "URLBROKER_PORT=6000",
                "DATASTORE_PORT=7000",
                "HEALTHCHECK_PORT=8000",
                "SCRAPER_PACKAGE=examplescraper",
                "SCRAPER_CONFIG=tests.integration.fake_config",
            ],
            "image": "dusanmadar/scrapemeagain:1.1.0",
            "volumes": ["/tmp/examplescraper:/scp/examplescraper"],
            "entrypoint": "/scrapemeagain/scrapemeagain/dockerized/entrypoints/entrypoint.scp1.sh",  # noqa
        },
        "examplescraper-scp2": {
            "environment": [
                "SERVICE_NAME_TEMPLATE=examplescraper-scp{}",
                "SERVICE_NAME_MASTER_SCRAPER=examplescraper-scp1",
                "DOCKER_HOST_IP=fake_docker_host_ip",
                "TOR_PORT=9051",
                "TOR_PASSWORD=I-solemnly-swear-I-am-up-to-no-good",
                "PRIVOXY_PORT=8118",
                "PRIVOXY_HOST=127.0.0.1",
                "IPSTORE_PORT=5000",
                "URLBROKER_PORT=6000",
                "DATASTORE_PORT=7000",
                "HEALTHCHECK_PORT=8000",
                "SCRAPER_PACKAGE=examplescraper",
                "SCRAPER_CONFIG=tests.integration.fake_config",
            ],
            "image": "dusanmadar/scrapemeagain:1.1.0",
            "volumes": ["/tmp/examplescraper:/scp/examplescraper"],
            "depends_on": ["examplescraper-scp1"],
            "entrypoint": "/scrapemeagain/scrapemeagain/dockerized/entrypoints/entrypoint.scpx.sh",  # noqa
        },
    },
}


class DockerComposeTestCase(unittest.TestCase):
    @patch("scrapemeagain-compose.DOCKER_HOST_IP", "fake_docker_host_ip")
    def test_construct_compose_dict(self):
        """
        Test `construct_compose_dict` returns expected compose dict.
        """
        self.assertEqual(
            EXPECTED_EXAMPLESCRAPER_COMPOSE_DICT,
            docker_compose.construct_compose_dict(
                "/tmp/examplescraper", "tests.integration.fake_config"
            ),
        )

    def test_construct_compose_dict_nonexisting_scraper(self):
        """
        Test `construct_compose_dict` raises `ModuleNotFoundError` for a
        nonexisting scraper.
        """
        with self.assertRaises(ModuleNotFoundError):
            docker_compose.construct_compose_dict("/nonexisting")



"""Common API definition for all scraper classes."""


import abc


class BaseScraper(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def base_url(self):
        """Target web site base address."""
        raise NotImplementedError()

    @abc.abstractproperty
    def list_url_template(self):
        """Target web site list page URL template."""
        raise NotImplementedError()

    @abc.abstractproperty
    def db_file(self):
        """SQLite file name."""
        raise NotImplementedError()

    @abc.abstractproperty
    def db_table(self):
        """SQLAlchemy table reference."""
        raise NotImplementedError()

    @abc.abstractmethod
    def generate_list_urls(self):
        """Generate list pages URLs.

        :returns iterator
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_lists_count(self):
        """Get lists count from the first list page pagination.

        :returns int
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_item_urls(self, response):
        """Get item URLs from a given list page.

        :argument response: list page
        :type response: `requests.Response`

        :returns
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_item_properties(self, response):
        """Get item properties.

        :argument response:
        :type response: `requests.Response`

        :returns dict
        """
        raise NotImplementedError()



import fcntl
import os
from subprocess import CalledProcessError, check_call, PIPE
import socket
import struct
from time import sleep

from flask import Flask

from scrapemeagain.config import Config
from scrapemeagain.utils.logger import setup_logging


def app_factory(name):
    """
    Create a Flask app using scraper specific config. Setup loging and add a
    basic `health` endpoint.

    To be able to use scraper specific config in a module this function has to
    be called first, e.g.

        app = app_factory(__name__)
        urlbroker_class = get_class_from_path(Config.URLBROKER_CLASS)

    :argument name: app name
    :type name: str

    :returns: Flask app
    """
    apply_scraper_config()

    app = Flask(name)

    logger_name = app.root_path.rsplit("/", 1)[-1]
    setup_logging(logger_name)

    @app.route("/health/")
    def healthcheck():
        return ""

    return app


def get_class_from_path(path):
    """
    Get class from given path.

    :argument path: dotted path to the target class
    :type path: str

    :returns: class
    """
    module_path, class_name = path.rsplit(".", 1)
    module = __import__(module_path, fromlist=[class_name])

    return getattr(module, class_name)


def apply_scraper_config(scraper_package=None, scraper_config=None):
    """
    Apply scraper specific config (by simply loading it's module).

    :argument scraper_package: scraper package name
    :type scraper_package: str
    :argument scraper_config: dotted path to scraper config
    :type scraper_config: str

    :returns: class
    """
    if scraper_config is None:
        scraper_config = os.environ.get("SCRAPER_CONFIG")

    if scraper_config is not None:
        get_class_from_path(scraper_config)
        return

    if scraper_package is None:
        scraper_package = os.environ.get("SCRAPER_PACKAGE")

    get_class_from_path("{}.config".format(scraper_package))


def scraper_is_running(hostname):
    """
    Check if a given scraper is running by trying to ping it.

    :argument hostname: scraper's hostname
    :type hostname: str

    :returns: bool
    """
    try:
        result = check_call(
            ["ping", "-c", "2", hostname], stdout=PIPE, stderr=PIPE
        )
    except CalledProcessError:
        result = 1

    # If ping's return code == 0 the scraper is running.
    return result == 0


def wait_for_other_scrapers():
    """
    Wait untill the rest of the scrapers is finished.
    """
    apply_scraper_config()

    for sraper_id in range(0, Config.SCRAPERS_COUNT):
        sraper_id += 1

        if sraper_id == 1:
            continue

        hostname = os.environ.get("SERVICE_NAME_TEMPLATE").format(sraper_id)
        while True:
            if not scraper_is_running(hostname):
                break

            sleep(2)


def get_inf_ip_address(ifname):
    """
    Get IP address of the given interface nameself.

    Credits https://stackoverflow.com/q/24196932/4183498.

    :argument ifname: interface name
    :type ifname: str

    :returns: str
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(
        fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack("256s", bytes(ifname[:15], "utf-8")),
        )[20:24]
    )


def inside_condainer():
    """
    Check if the process runs inside a container.

    Credits: https://stackoverflow.com/q/23513045/4183498.

    :returns: bool
    """
    return os.path.exists("/.dockerenv")



import os

from requests import get

from scrapemeagain.config import Config


def check_ip_safeness(ip):
    """
    Helper to check IP safeness.

    :argument ip: current Tor IP
    :type ip: str

    :returns bool
    """
    url = "http://{host}:{port}/ip-is-safe/{ip}/".format(
        host=os.environ.get("SERVICE_NAME_MASTER_SCRAPER"),
        port=Config.IPSTORE_PORT,
        ip=ip,
    )
    return get(url).json()["safe"]



"""
Common database functionality.
"""


import logging
import os
import socket

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from scrapemeagain.config import Config
from scrapemeagain.dockerized.apps.datastore import client as datastore_client
from scrapemeagain.scrapers.basemodel import ItemUrlsTable


class BaseDatabaser:
    def __init__(self, db_name, data_table):
        """
        Database manipulation class.

        :argument db_name: SQLite database file name (without extension)
        :type db_name: str
        :argument data_table: table to store item data in
        :type data_table: SQLAlchemy table
        """
        self.db_name = db_name

        self.item_data_table = data_table
        self.item_urls_table = ItemUrlsTable

        self.transaction_items = 0
        self.transaction_items_max = Config.TRANSACTION_SIZE

        self.engine = self.create_engine()
        self.session = sessionmaker(bind=self.engine)()

    def create_engine(self):
        """
        Create an SQLite database engine.
        """
        # Ensure data dir exists.
        if not os.path.exists(Config.DATA_DIRECTORY):
            os.makedirs(Config.DATA_DIRECTORY)

        db = os.path.join(Config.DATA_DIRECTORY, self.db_name) + ".sqlite"
        if not os.path.exists(db):
            with open(db, "w"):
                pass

        return create_engine("sqlite:///{}".format(db))

    def create_tables(self, create_urls_table=True, create_data_table=True):
        """
        Create tables.
        """
        if create_urls_table:
            self.item_urls_table.__table__.create(self.engine, checkfirst=True)

        if create_data_table:
            self.item_data_table.__table__.create(self.engine, checkfirst=True)

    def commit(self):
        """
        Commit changes.
        """
        try:
            self.session.commit()
            self.transaction_items = 0
            logging.info("Changes successfully committed")
        except Exception as exc:
            logging.error("Failed to commit changes, rolling back ...")
            logging.exception(exc)
            self.session.rollback()

    def manage_transaction(self):
        """
        Manage transaction, i.e. decide when to commit.
        """
        self.transaction_items += 1
        if self.transaction_items > self.transaction_items_max:
            self.commit()

    def _actually_insert(self, data, table):
        """
        Insert data to table.

        :argument data:
        :type data: dict
        :argument table: table reference
        :type table: SQLAlchemy table
        """
        row = table()
        for key, value in data.items():
            setattr(row, key, value)

        self.session.add(row)

    def insert(self, data, table):
        """
        Insert a single data unit and manage the transaction size.
        """
        self._actually_insert(data, table)
        self.manage_transaction()

    def insert_multiple(self, data, table):
        """
        Insert multiple data units.

        :argument data:
        :type data: list of dicts
        :argument table: table reference
        :type table: SQLAlchemy table
        """
        for item in data:
            self.insert(item, table)

    def delete_url(self, url):
        """
        Delete item URL.

        :argument url:
        :type url: str
        """
        self.session.query(self.item_urls_table).filter(
            self.item_urls_table.url == url
        ).delete()

        self.transaction_items += 1

    def _remove_duplicate_item_urls(self):
        """
        Remove duplicate item URLs.
        """
        raw_sql = """
            DELETE
            FROM {table}
            WHERE {id} IN (
                SELECT MAX({id})
                FROM {table}
                GROUP BY {url}
                HAVING COUNT(*) > 1
            )
        """.format(
            id=self.item_urls_table.id.key,
            url=self.item_urls_table.url.key,
            table=self.item_urls_table.__tablename__,
        ).strip()

        self.engine.execute(raw_sql)

    def get_item_urls(self):
        """
        Get item URLs for scraping ordered from newest to oldest.

        :returns query object
        """
        self._remove_duplicate_item_urls()

        return self.session.query(self.item_urls_table.url).order_by(
            self.item_urls_table.id.desc()
        )


class Databaser(BaseDatabaser):
    def __init__(self, db_name, data_table):
        super().__init__(db_name, data_table)
        self.create_tables()


class DataOnlyDatabaser(BaseDatabaser):
    """
    Store item data only.

    Instance of this class will have tables set as follows:
        self.item_data_table = data_table
        self.item_urls_table = None
    """

    def __init__(self, db_name, data_table):
        super().__init__(db_name, data_table)
        self.item_urls_table = None
        self.create_tables(create_urls_table=False)

    def insert(self, data):
        super().insert(data, self.item_data_table)


class UrlsOnlyDatabaser(BaseDatabaser):
    """
    Store item URLs only.

    Instance of this class will have tables set as follows:
        self.item_data_table = None
        self.item_urls_table = ItemUrlsTable
    """

    def __init__(self, db_name):
        super().__init__(db_name, None)
        self.create_tables(create_data_table=False)


class DataStoreDatabaser(DataOnlyDatabaser):
    """
    A data only databaser intended to used as the `datastore` app DB backend.
    """

    def deserialize_data(self, data):
        """
        Update the serialized `data` dict to match the DB schema and return it.
        """
        return data

    def insert(self, data):
        data = self.deserialize_data(data)
        super().insert(data)


class DockerizedDatabaser(UrlsOnlyDatabaser):
    """
    A hybrid Databaser which stores item URLs locally but item data remotely.

    This is the databaser class each dockerized scraped should use/subclass.
    """

    def __init__(self, db_name):
        super().__init__("{0}_{1}".format(db_name, socket.gethostname()))

    def serialize_data(self, data):
        """
        Update the raw `data` dict to be JSON serializable and return it.
        """
        return data

    def insert(self, data, table):
        if table is None:
            # Store item data remotely
            # (`self.item_data_table = None` as we are a `UrlsOnlyDatabaser`).
            data = self.serialize_data(data)
            datastore_client.insert_data(data)
        else:
            # Store item URLs locally.
            super()._actually_insert(data, table)

        self.manage_transaction()

    def commit(self):
        datastore_client.commit()
        super().commit()



#!/usr/bin/env python3


"""
Dynamic `docker-compose.yml` based on settings defined in `config`.

Usage:
    `python3 scrapemeagain-compose.py -s <scraper> [-c <path.to.some.config>] | docker-compose -f - up`

    Example:
    $ cd ../
    $ python3 scrapemeagain-compose.py -s $(pwd)/examples/examplescraper | docker-compose -f - up
    $ python3 scrapemeagain-compose.py -s $(pwd)/examples/examplescraper -c tests.integration.fake_config | docker-compose -f - up
"""  # noqa


import argparse
import os
import sys

import yaml

from scrapemeagain.config import Config
from scrapemeagain.dockerized.utils import (
    apply_scraper_config,
    get_inf_ip_address,
)

# Must match Dockerfile's `SCP_DIR` env variable.
SCP_DIR = "/scp"
# Must match Dockerfile's `APP_SRC_DIR` env variable.
APP_SRC_DIR = "/scrapemeagain/scrapemeagain"

ENTRYPOINT_PATH_TEMPLATE = (
    APP_SRC_DIR + "/dockerized/entrypoints/entrypoint.{}.sh"
)


DOCKER_HOST_IP = get_inf_ip_address(Config.DOCKER_INTERFACE_NAME)


def create_scraper_service(
    sraper_id, scraper_package, scraper_path, scraper_config
):
    service_name_template = "{0}-scp".format(scraper_package)
    service_name_template += "{}"
    # The first scraper service is always the master one.
    service_name_master = service_name_template.format(1)
    service_name = service_name_template.format(sraper_id)

    service_settings = {
        "environment": [
            "SERVICE_NAME_TEMPLATE={}".format(service_name_template),
            "SERVICE_NAME_MASTER_SCRAPER={}".format(service_name_master),
            "DOCKER_HOST_IP={}".format(DOCKER_HOST_IP),
            "TOR_PORT={}".format(Config.TOR_PORT),
            "TOR_PASSWORD={}".format(Config.TOR_PASSWORD),
            "PRIVOXY_PORT={}".format(Config.PRIVOXY_PORT),
            "PRIVOXY_HOST={}".format(Config.PRIVOXY_HOST),
            "IPSTORE_PORT={}".format(Config.IPSTORE_PORT),
            "URLBROKER_PORT={}".format(Config.URLBROKER_PORT),
            "DATASTORE_PORT={}".format(Config.DATASTORE_PORT),
            "HEALTHCHECK_PORT={}".format(Config.HEALTHCHECK_PORT),
            "SCRAPER_PACKAGE={}".format(scraper_package),
        ],
        "image": "dusanmadar/scrapemeagain:1.1.0",
        "volumes": [
            "{}:{}".format(
                scraper_path, os.path.join(SCP_DIR, scraper_package)
            )
        ],
    }

    if scraper_config is not None:
        service_settings["environment"].append(
            "SCRAPER_CONFIG={}".format(scraper_config)
        )

    if sraper_id == 1:
        # Master scraper specific settings.
        entrypoint = ENTRYPOINT_PATH_TEMPLATE.format("scp1")
    else:
        # Worker specific settings.
        entrypoint = ENTRYPOINT_PATH_TEMPLATE.format("scpx")
        service_settings["depends_on"] = [
            service_name_template.format(sraper_id - 1)
        ]

    service_settings["entrypoint"] = entrypoint

    return {service_name: service_settings}


def construct_compose_dict(scraper_path, scraper_config=None):
    # Apply scraper specific config.
    scraper_package = scraper_path.split(os.sep)[-1]
    prepare_for_apply_scraper_config(scraper_path)
    apply_scraper_config(scraper_package, scraper_config)

    docker_compose = {"version": "3", "services": {}}

    for sraper_id in range(0, Config.SCRAPERS_COUNT):
        sraper_id += 1

        docker_compose["services"].update(
            create_scraper_service(
                sraper_id, scraper_package, scraper_path, scraper_config
            )
        )

    return docker_compose


def prepare_for_apply_scraper_config(scraper_path):
    sys.path.insert(0, os.path.dirname(scraper_path))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s",
        "--scraper",
        required=True,
        help=(
            "Absolute path to scraper package dir; e.g. '/tmp/examplescraper'."
        ),
    )
    parser.add_argument(
        "-c",
        "--config",
        required=False,
        help=(
            "Dotted path to a specific (not 'config.py') config module; "
            "e.g. 'examplescraper.more_configs.test_config'. The path must be"
            "importbale locally and in the container."
        ),
        default=None,
    )

    args = parser.parse_args()
    print(yaml.dump(construct_compose_dict(args.scraper, args.config)))


if __name__ == "__main__":
    main()



from toripchanger import TorIpChanger

from scrapemeagain.dockerized.apps.ipstore.client import check_ip_safeness


class DockerizedTorIpChanger(TorIpChanger):
    def _ip_is_safe(self, current_ip):
        return check_ip_safeness(current_ip)

    def _manage_used_ips(self, current_ip):
        # No need to maintain used IPs locally the global IP store takes care
        # of that.
        pass



from datetime import datetime
import json
import multiprocessing
import os
import subprocess
import tempfile
from time import sleep
import unittest
import yaml

from sqlalchemy import create_engine
from scrapemeagain.dockerized.utils import get_inf_ip_address
from scrapemeagain.utils.alnum import DATE_FORMAT

from examplescraper.config import Config
from examplescraper.examplesite.app import app as examplesite_app
from tests import REPO_DIR, EXAMPLESCRAPER_DIR
from tests.utils import import_docker_compose


docker_compose = import_docker_compose()


DOCKER_HOST_IP = get_inf_ip_address(Config.DOCKER_INTERFACE_NAME)

USER_AGENTS_FILE = os.path.join(REPO_DIR, "useragents.json")
USER_AGENTS_FILE_BACKUP = USER_AGENTS_FILE + ".bak"
FAKE_USER_AGENTS_JSON = {
    "date": datetime.today().strftime(DATE_FORMAT),
    "useragents": ["User Agent 1", "User Agent 2", "User Agent 3"],
}


class IntegrationTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Prevent populating useragents file.
        if os.path.exists(USER_AGENTS_FILE):
            os.rename(USER_AGENTS_FILE, USER_AGENTS_FILE_BACKUP)

        if not os.path.exists(USER_AGENTS_FILE_BACKUP):
            with open(USER_AGENTS_FILE, "w") as outfile:
                json.dump(FAKE_USER_AGENTS_JSON, outfile)

    @classmethod
    def tearDownClass(cls):
        if os.path.exists(USER_AGENTS_FILE_BACKUP):
            os.rename(USER_AGENTS_FILE_BACKUP, USER_AGENTS_FILE)

    def _run_examplesite(self):
        p = multiprocessing.Process(
            target=examplesite_app.run, args=(DOCKER_HOST_IP, 9090)
        )
        p.start()

        return p

    def _run_examplescraper_compose(self):
        scrapemeagain_compose = yaml.dump(
            docker_compose.construct_compose_dict(
                EXAMPLESCRAPER_DIR, "tests.integration.fake_config"
            )
        )
        # ' - ' meaning: https://unix.stackexchange.com/a/16364/266262.
        command = "docker-compose -f - up --force-recreate"
        p = subprocess.Popen(command.split(" "), stdin=subprocess.PIPE)
        p.communicate(input=bytes(scrapemeagain_compose, "utf-8"))

    def _get_container_name(self, name_pattern, queue):
        def get_running_container_name(name_pattern, queue):
            command = "docker ps | grep {}".format(name_pattern)
            scraper_name = None

            while scraper_name is None:
                try:
                    out = subprocess.check_output(command, shell=True)
                    out = out.strip()
                    scraper_name = out.split()[-1]
                except subprocess.CalledProcessError:
                    pass

                sleep(1)

            queue.put(scraper_name)

        p = multiprocessing.Process(
            target=get_running_container_name, args=(name_pattern, queue)
        )
        p.start()

        return p

    def _copy_scraper_files(self, container_name, tmp_dir_path):
        command = f"docker cp {container_name}:/tmp/test-examplescraper {tmp_dir_path}"  # noqa
        subprocess.check_call(command.split(" "))

    def validate_db(self, db_path):
        self.assertTrue(os.path.exists(db_path))

        engine = create_engine(f"sqlite:///{db_path}")

        expected_table_name = "example_item_data"
        self.assertIn(expected_table_name, engine.table_names())

        sql = f"SELECT count(*) FROM {expected_table_name}"
        # 100 = 10 list pages * 10 post links per list page.
        self.assertEqual(engine.execute(sql).fetchone()[0], 100)

    def test_examplescraper_works(self):
        """
        Test `examplescraper` can scrape required data from `examplesite`.
        """
        examplesite = self._run_examplesite()
        # TODO there has to a better way to wait till the process has started.
        sleep(1)
        if not examplesite.is_alive():
            print("Failed to start examplesite")
            return

        container_name_queue = multiprocessing.Queue()
        container_name_getter = self._get_container_name(
            "examplescraper-scp1", container_name_queue
        )

        try:
            self._run_examplescraper_compose()
        finally:
            examplesite.terminate()
            examplesite.join()

            container_name_getter.terminate()
            container_name_getter.join()

        with tempfile.TemporaryDirectory() as tmp_dir_path:
            container_name = container_name_queue.get().decode("utf-8")
            self._copy_scraper_files(container_name, tmp_dir_path)

            expected_db_file = os.path.join(
                tmp_dir_path, "test-examplescraper", "example.sqlite"
            )
            self.validate_db(expected_db_file)



"""Backbone services management."""


from getpass import getpass
import subprocess
import sys


SERVICES = ("privoxy", "tor")
COMMAND = ("sudo", "-S", "service")

password_manager = {"root_password": ""}


def toggle_backbone_services(action):
    """
    Start/stop backbone services required by the scraper.

    :argument action: 'start' or 'stop'
    :type string_ts: str
    """
    if not password_manager["root_password"]:
        password_manager["root_password"] = getpass().strip()

    for service in SERVICES:
        command = list(COMMAND)
        command.extend((service, action))

        service_action = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        service_action.communicate(
            (password_manager["root_password"] + "\n").encode()
        )
        service_action.wait()

        if service_action.returncode != 0:
            sys.exit("Failed to {0} service {1}".format(action, service))


def start_backbone_services():
    toggle_backbone_services("start")


def stop_backbone_services():
    toggle_backbone_services("stop")



"""ScrapeMeAgain Package Handler"""


from setuptools import setup, find_packages


description = "Yet another Python web scraping application"


setup(
    version="1.0.1",
    name="scrapemeagain",
    url="https://github.com/DusanMadar/ScrapeMeAgain",
    author="Dusan Madar",
    author_email="madar.dusan@gmail.com",
    description=description,
    long_description=description,
    keywords="web scraper",
    packages=find_packages(),
    scripts=["scripts/scrapemeagain-compose.py"],
    test_suite="tests",
    license="MIT",
    platforms="linux",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Internet :: WWW/HTTP",
    ],
)



from scrapemeagain.config import Config


class UrlsRangeManager:
    def __init__(self, descending=True):
        self._urls_range = None
        self.descending = descending

    def get_urls_count(self):
        return 0

    def generate_urls_range(self):
        urls_count = self.get_urls_count()
        step = urls_count // Config.SCRAPERS_COUNT

        for i in range(urls_count, 0, -step):
            start = i
            end = i - step
            if end < step:
                end = 0

            # The returned range is expected to be processed in descending
            # order: start > end, e.g. (100, 0); unless the `descending`
            # flag is unset: start < end, e.g. (0, 100).
            yield (start, end) if self.descending else (end, start)

            if end == 0:
                break

    def get_urls_range(self):
        if self._urls_range is None:
            self._urls_range = self.generate_urls_range()

        try:
            return next(self._urls_range)
        except StopIteration:
            return (0, 0)


class ListUrlsBroker(UrlsRangeManager):
    def __init__(self, scraper):
        super().__init__()

        self.scraper = scraper

    def get_urls_count(self):
        return self.scraper.get_lists_count()



import os

from requests import get, post

from scrapemeagain.config import Config


def _get_datastore_url(endpoint):
    return "http://{host}:{port}/{endpoint}/".format(
        host=os.environ.get("SERVICE_NAME_MASTER_SCRAPER"),
        port=Config.DATASTORE_PORT,
        endpoint=endpoint,
    )


def insert_data(data):
    url = _get_datastore_url("insert-data")
    post(url, json=data)


def commit():
    url = _get_datastore_url("commit")
    get(url)



import os

from requests import get

from scrapemeagain.config import Config


def _get_urls_range(endpoint):
    """
    Get list URLs range.

    :returns tuple
    """
    url = "http://{host}:{port}/{endpoint}/".format(
        host=os.environ.get("SERVICE_NAME_MASTER_SCRAPER"),
        port=Config.URLBROKER_PORT,
        endpoint=endpoint,
    )
    response_json = get(url).json()

    return (response_json["start"], response_json["end"])


def get_list_urls_range():
    return _get_urls_range("list-urls-range")


def get_item_urls_range():
    return _get_urls_range("item-urls-range")
